// Copyright (c) 2010-2016 The Regents of the University of Michigan
// This file is part of the Freud project, released under the BSD 3-Clause License.

#include <memory>

#include <algorithm>
#include "LinkCell.h"
// hack to keep VectorMath's swap from polluting the global namespace
// if this is a problem, we need to solve it
#include "VectorMath.h"
#include "box.h"
#include "Index1D.h"

#include "tbb/atomic.h"

#ifndef _NEAREST_NEIGHBORS_H__
#define _NEAREST_NEIGHBORS_H__

/*! \file NearestNeighbors.h
  \brief Find the requested number of nearest neighbors
*/

namespace freud { namespace locality {

//! Iterates over particles in a neighbor list generated by NearestNeighbor
/*! While the neighbor list is simple to iterate over, it is not efficient to do so at the C++ level in a threaded
    environment. This helper class makes that easier both in c++ and
    provides a python compatibile interface for direct usage there.

    An IteratorNeighborList is given the bare essentials it needs to iterate over a neighbor list.
    Call begin() to get the fist neighbor index, next() to get the index of the next
    particle in the cell, atEnd() will return true if you are at the end. In C++, next() will crash the code if you
    attempt to iterate past the end (no bounds checking for performance). When called from python, a different version
    of next is used that will throw StopIteration at the end.

    A loop over all of the particles in a cell can be acomplished with the following code in c++.
\code
 locality::NearestNeighbors::iteratorneighbor it = nn.iterneighbor(i);
 for (unsigned int j = it.begin(); !it.atEnd(); j = it.next())
     {
     // do something with particle i
     }
\endcode

\note Behavior is undefined if an IteratorNeighborList is accessed after the parent LinkCell is destroyed.
*/
class IteratorNeighborList
    {
    public:
        IteratorNeighborList(const std::shared_ptr<unsigned int>& neighbor_list,
                             unsigned int n,
                             unsigned int k,
                             unsigned int idx)
                             : m_neighbor_list(neighbor_list.get()), m_n(n), m_k(k), m_idx(idx)
            {
            m_cur_idx = 0;
            }

        //! Test if the iteration over the cell is complete
        bool atEnd()
            {
            return (m_cur_idx == m_k);
            }

        //! Get the next particle index in the list
        unsigned int next()
            {
            m_cur_idx++;
            return m_neighbor_list[m_idx*m_k + m_cur_idx];
            }

        //! Get the first particle index in the list
        unsigned int begin()
            {
            m_cur_idx = 0;
            return m_neighbor_list[m_idx*m_k + m_cur_idx];
            }

    private:
        const unsigned int *m_neighbor_list;                  //!< The neighbor list
        unsigned int m_k;                                //!< Number of neighbors in list
        unsigned int m_n;                                //!< Number of particles in list
        unsigned int m_idx;                                //!< Index of particle being iterated
        unsigned int m_cur_idx;                           //!< Current index
    };

/*! Find the requested number of nearest neighbors
*/
class NearestNeighbors
    {
    public:
        //! iterator to iterate over neighbors
        typedef IteratorNeighborList iteratorneighbor;
        // Null constructor for use in triclinic; will be removed when cell list is fixed
        NearestNeighbors();
        //! Constructor
        NearestNeighbors(float rmax,
                         unsigned int num_neighbors,
                         float scale=1.1,
                         bool strict_cut=false);

        ~NearestNeighbors();

        //! Iterate over particles in a cell
        iteratorneighbor iterneighbor(unsigned int idx) const
            {
            assert(m_neighbor_array.get() != NULL);
            return iteratorneighbor(m_neighbor_array, m_num_ref, m_num_neighbors, idx);
            }

        void setRMax(float rmax)
            {
            m_rmax = rmax;
            m_lc->setCellWidth(m_rmax);
            }

        void setRMaxPy(float rmax)
            {
            m_rmax = rmax;
            m_lc->setCellWidth(m_rmax);
            }

        //! Get the simulation box
        const box::Box& getBox() const
            {
            return m_box;
            }

        //! Get the val for UINTMAX
        unsigned int getUINTMAX() const
            {
            return UINT_MAX;
            }

        //! Get the number of neighbors
        unsigned int getNumNeighbors() const
            {
            return m_num_neighbors;
            }

        //! Get the number of reference points we've computed for
        unsigned int getNref() const
            {
            return m_num_ref;
            }

        //! Get the number of particles we've computed for
        unsigned int getNp() const
            {
            return m_num_points;
            }

        //! Get the current cutoff radius used
        float getRMax() const
            {
            return m_rmax;
            }

        //! Get the current cutoff radius used
        float getRMaxPy() const
            {
            return m_rmax;
            }

        //! Get a reference to the neighbors array
        std::shared_ptr<unsigned int> getNeighbors(unsigned int i) const
            {
            // create the array
            std::shared_ptr<unsigned int> requested_neighbors = std::shared_ptr<unsigned int>(new unsigned int[m_num_neighbors], std::default_delete<unsigned int[]>());
            // find the position from which to read neighbors
            unsigned int start_idx = i*m_num_neighbors;
            for (unsigned int j=0; j<m_num_neighbors; j++)
                {
                requested_neighbors.get()[j] = m_neighbor_array.get()[start_idx + j];
                }
            return requested_neighbors;
            }

        //! Get a reference to the neighborlist array
        std::shared_ptr<unsigned int> getNeighborList() const
            {
            return m_neighbor_array;
            }

        //! Get a reference to the distance array
        std::shared_ptr<float> getRsq(float i) const
            {
            // create the array
            std::shared_ptr<float> requested_rsq = std::shared_ptr<float>(new float[m_num_neighbors], std::default_delete<float[]>());
            // find the position from which to read neighbors
            unsigned int start_idx = i*m_num_neighbors;
            for (unsigned int j=0; j<m_num_neighbors; j++)
                {
                requested_rsq.get()[j] = m_rsq_array.get()[start_idx + j];
                }
            return requested_rsq;
            }

        //! Get a reference to the distanceList array
        std::shared_ptr<float> getRsqList() const
            {
            return m_rsq_array;
            }

        //! Get a reference to the wrapped vector array
        std::shared_ptr<vec3<float> > getWrappedVectors() const
            {
            return m_wvec_array;
            }

        void setCutMode(const bool strict_cut);

        //! find the requested nearest neighbors
        void compute(const box::Box& box, const vec3<float> *ref_pos, unsigned int n_ref, const vec3<float> *pos, unsigned int Np);

    private:
        box::Box m_box;            //!< Simulation box the particles belong in
        float m_rmax;                     //!< Maximum r at which to determine neighbors
        unsigned int m_num_neighbors;            //!< Number of neighbors to calculate
        float m_scale;                    //!< scale by which to increase neighbor search radius
        bool m_strict_cut;                  //!< use a strict r_cut, or allow freud to expand the r_cut as needed
        unsigned int m_num_points;                //!< Number of particles for which nearest neighbors checks
        unsigned int m_num_ref;                //!< Number of particles for which nearest neighbors calcs
        locality::LinkCell* m_lc;          //!< LinkCell to bin particles for the computation
        tbb::atomic<unsigned int> m_deficits; //!< Neighbor deficit count from the last compute step
        std::shared_ptr<unsigned int> m_neighbor_array;         //!< array of nearest neighbors computed
        std::shared_ptr<float> m_rsq_array;         //!< array of distances to neighbors
        std::shared_ptr<vec3<float> > m_wvec_array;         //!< array of distances to neighbors
        };

}; }; // end namespace freud::locality

#endif // _NEAREST_NEIGHBORS_H__
